#
# camera/formats/config.py
#
# Camera Configuration File Reader
#
# This file is part of the Python package to handle Camera data (CAMPP).
#
# Copyright (c) 2011-2012, Universiteit Leiden
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# - Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# - Neither the name of Universiteit Leiden nor the names of its contributors
#   may be used to endorse or promote products derived from this software
#   without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
#
# Authors: Sander Roobol
#
# $Id: config.py 649 2012-02-27 13:41:59Z moene $
#

# TODO:
# - read digital scan and approach settings

"""This module provides read access to Camera Configuration files (CCameraDeviceConfig dumps).

Standalone usage:
python -m camera.formats.config path/to/file.config
"""


from . import util

class VersionError(Exception):
    pass

class ConfigReader(object):
    _fields = [
        ('ConfigFileVersion',    util.INT32), # required to be 0x20090221
        ('NewClassTag',          util.INT16), # 0xFFFF excpected, as it is this object's first occurrence in archive
        ('ConfigSchemaVersion',  util.INT16), # code supports version <= 7 but this is not strictly enforced
        ('ConfigObjectName',     util.OBJNAME),
        ('ScanAreaSizeTrian',    util.UINT32),
        ('ScanAreaSizeTooth',    util.UINT32),
        ('ScanRotation',         util.DOUBLE),
        ('ScanAreaStatOffsetX',  util.INT32),
        ('ScanAreaStatOffsetY',  util.INT32),
        ('ScanAreaAspectRatioX', util.UINT32),
        ('ScanAreaAspectRatioX', util.UINT32),
        ('PixelClock',           util.DOUBLE),
        ('SampleRate',           util.UINT32),
        ('SizeTrian',            util.UINT32),
        ('SizeTooth',            util.UINT32),
        ('OverSampling',         util.UINT32),
        ('XYConversion',         util.DOUBLE),
        ('ADCConversion',        util.DOUBLE),
        ('FullFrameUpdate',      util.INT32),
        ('XRounding',            util.UINT32),
        ('BackStroke',           util.UINT32),
        ('BandWidth',            util.UINT32),
        ('ZOffset',              util.DOUBLE),
        ('ZOffsetAd',            util.DOUBLE),
        ('AutoZOffset',          util.INT32),
        ('ZdZFactor',            util.DOUBLE),
    ] + sum([[
            ('ADC{0} enabled'.format(i),    util.INT32),
            ('ADC{0} gain'.format(i),       util.UINT32),
            ('ADC{0} conversion'.format(i), util.DOUBLE),
            ('ADC{0} name'.format(i),       util.CSTR),
            ('ADC{0} unit'.format(i),       util.CSTR),
    ] for i in range(8)], []) + [
        ('FeedBack',                    util.INT32),
        ('IntegrationTime',             util.UINT32),
        ('ProportionalGain',            util.UINT32),
        ('WishedTunnelingCurrent',      util.UINT32),
        ('SampleVoltage',               util.DOUBLE),
        ('VoltagePulseDuration',        util.UINT32),
        ('VoltagePulseHeight',          util.INT32),
        ('TipCrashDepth',               util.UINT32),
        ('BiasVoltageBandwidth_Hz',     util.UINT32), # schema version 6/7 only
        ('ExtraMixerOffset',            util.DOUBLE),
        ('ExtraMixerAmpX',              util.INT32),
        ('ExtraMixerAmpY',              util.INT32),
        ('ExtraMixerAttenuation',       util.UINT32),
        ('ExtraMixerSlewRate',          util.UINT32),
        ('3dMeasPerPixel',              util.INT32),
        ('3dDisplayIndex',              util.INT32),
        ('3dFrameNumber',               util.INT32),
        ('MaxTipSpeed_nmS',             util.UINT32),
        ('MaxScanAreaSize_nm',          util.UINT32),
        ('MinStepTime',                 util.UINT32),
        ('WaveFormGeneratorDelayAfter', util.UINT32),
        ('ApproachSampleVoltage',       util.DOUBLE),
        ('AvgOverNSamples',             util.UINT32),
        ('NStep',                       util.UINT32),
        ('Sleep1',                      util.UINT32),
        ('Sleep2',                      util.UINT32),
        ('ApproachGain',                util.UINT32),
        ('ApproachThreshold',           util.DOUBLE),
        ('COMport',                     util.UINT32),
        ('EnableDialsbox',              util.INT32),
    ]

    def __init__(self, filename):
        self.filename = filename
        with open(filename, 'rb') as fp:
            self._read(fp)

    def _read(self, fp):
        self.__dict__.update(util.readBinaryHeader(fp, self._fields[:3]))
        if self.ConfigFileVersion != 0x20090221:
            raise VersionError("unsupported file version 0x{0:x}".format(self.ConfigFileVersion))
        if self.ConfigSchemaVersion < 6:
            # make a copy of self._fields that does not contain the BiasVoltageBandwidth_Hz field
            # it needs to be a copy since self._fields is shared by all instances because it's defined at the class level
            self._fields = [(name, type) for (name, type) in self._fields if name != 'BiasVoltageBandwidth_Hz']
        self.__dict__.update(util.readBinaryHeader(fp, self._fields[3:]))

    def __str__(self):
        # I apologise for this ugly non-pythonic line of code
        return "ConfigReader {0}:\n  ConfigFileVersion: 0x{1:x}\n{2}".format(self.filename, self.ConfigFileVersion, "\n".join("  {0}: {1}".format(name, getattr(self, name)) for (name, type) in self._fields[1:]))

if __name__ == '__main__':
    import sys

    if len(sys.argv) != 2:
        print("usage: {0} file.config".format(sys.argv[0]))
        sys.exit(1)

    info = ConfigReader(sys.argv[1])
    print(info)
