#
# camera/formats/raw.py
#
# Camera Raw File Reader
#
# This file is part of the Python package to handle Camera data (CAMPP).
#
# Copyright (c) 2009-2012, Universiteit Leiden
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# - Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# - Neither the name of Universiteit Leiden nor the names of its contributors
#   may be used to endorse or promote products derived from this software
#   without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
#
# Authors: Martin Moene (original), Sander Roobol
#
# $Id: raw.py 649 2012-02-27 13:41:59Z moene $
#

#
# ToDo:
# - handling of different file format versions
# - make RawXXX dictionary based like RawFileHeader ? OR
# - make RawFileHeader non-dictionary based like RawXXX ?
#
# Notes:
# - properties require new-style classes to work correctly, see
#   http://stackoverflow.com/questions/598077/why-does-foo-setter-in-python-not-work-for-me
#   this can be enforced by deriving from object: class MyClass(object), or by
#   defining __metaclass__.
#

"""Module raw provides access to frame and channel information of Camera raw files.

Camera is the control and analysis application for the Leiden Probe Microscopy
Scanning Tunneling Microscope (STM) and for the Atomic Force Microscope (AFM).
Camera is the acronym for Computer Aided Measurement Environment for Realtime
Atomic imaging.

During a measurement Camera collects the data in a 'swap file'. When the measurement
has been completed, Camera creates a 'raw file' from the data in the named swap file.

Raw File Format

   The raw file contains information about the measurements and the image data itself.
   This information is roughly organized as follows:

                           (class RawFileReader)
      .-------------------------------------------------.
      | Header             (class RawFileHeader)        |
      |-------------------------------------------------|
      | Frame 0 info       (class RawFileFrameInfo)     |
      |   Channel 0 info   (class RawFileChannelInfo    |
      |   Channel 0 image  (class RawFileChannelImage)  |
      |          ...                                    |
      |   Channel K info                                |
      |   Channel K image                               |
      |-------------------------------------------------|
      |                     ...                         |
      |-------------------------------------------------|
      | Frame N info                                    |
      |   Channel 0 info                                |
      |   Channel 0 image                               |
      |          ...                                    |
      |   Channel K info                                |
      |   Channel K image                               |
      '-------------------------------------------------'

Header

   The file contains a single header with information such as hardware interface
   id and version and channels used and their names. For example:

      header: file version id: RAW1305
      header: interface id: CameraSTM_d
      header: interface version: 4.0
      header: frame count: 434
      header: channel mask: 0x7, 0b111
      header: channel count: 3
      header: channel0 name: Topographical height map
      header: channel1 name: Topographical height map [dZ]
      header: channel2 name:
      header: channel3 name:
      header: channel4 name:
      header: channel5 name:
      header: channel6 name:
      header: channel7 name:
      header: channel0 unit: nm
      header: channel1 unit: nm
      header: channel2 unit:
      header: channel3 unit:
      header: channel4 unit:
      header: channel5 unit:
      header: channel6 unit:
      header: channel7 unit:

   The header is represented by class RawFileHeader.

Frame Information

   For each frame there is a section with information about the measurement, such
   as the acquisition time, the scanned physical area (RawFileFrameScanArea) and
   the measurement (device) settings (RawFileFrameDeviceInfo). For example:

      frame 0: id: 0
      frame 0: acquisition time: 1275418042, Tue Jun 01 18:47:22 2010
      frame 0: x-offset [pm]: 35000
      frame 0: y-offset [pm]: 95000
      frame 0: x-step [pm]: 19.53125
      frame 0: x-step [pm]: 19.53125
      frame 0: x-size [pixels]: 512
      frame 0: y-size [pixels]: 512
      frame 0: scan angle [rad]: 0.0
      frame 0: sizeof device info: 440
      frame 0: pixel clock [kHz]: 80.0
      frame 0: sample per point: 8
      frame 0: oversampling factor: 8
      frame 0: z-gain: 32
      frame 0: dz-gain: 1
      frame 0: bandwidth [kHz]: 600
      frame 0: z-offset [??]: -1.85
      frame 0: zoverdz: 1.0
      frame 0: x-rounding [%]: 0.0
      frame 0: backwardstroke: 1024
      frame 0: u-sample [V]: 2.0
      frame 0: triangle attenuation: 0, 0x0
      frame 0: sawtooth attenuation: 0, 0x0
      frame 0: x-sample slope: 0
      frame 0: y-sample slope: 0
      frame 0: integration time code: 112
      frame 0: proportional gain code: 6
      frame 0: setvalue code: 46
      frame 0: 3D measurements per pixel: 0
      frame 0: 3D display index: 0
      frame 0: 3D frame number: 0
      frame 0: extra mixer offset [V]: 0.0
      frame 0: extra mixer x-gain: 0
      frame 0: extra mixer y-gain: 0
      frame 0: extra mixer attenuation: 4
      frame 0: extra mixer slewrate: 65535
      frame 0: adc channels: array('H', [0, 0, 0, 0, 42076, 271, 0, 0])
      frame 0: iv name:

   The frame information is represented by class RawFileFrameInfo.

   Please note that the frames are selected by their sequence number in the file,
   not by their frameId; these may differ if frames have been deleted.

Channel Information

   For each channel in a frame there is information about the size of the image
   and of the scaling of its data. For example frame 0, channel 0:

      channel 0.0: flags: 3, 0b11
      channel 0.0: x-size: 512
      channel 0.0: y-size: 513
      channel 0.0: pixels: 525312
      channel 0.0: data scale factor: 0.00074065625
      channel 0.0: data offset: -193986
      channel 0.0: parameter mask: 7, 0b111
      channel 0.0: x-slope [?]: 0.0
      channel 0.0: y-slope [?]: 0.0
      channel 0.0: average [?]: 162.075654087
      channel 0.0: z-min: -32768
      channel 0.0: z-max: 32766

   The channel information is represented by class RawFileChannelInfo.

Channel Images

   And finally for each channel in a frame there is the image information
   (16-bit signed integers), represented by class RawFileChannelImage.


Classes, Exceptions and Functions

Classes
   - RawFileReader - provide access to the various parts of a Camera raw file
   - RawFileHeader - provide access to the elements of the file header
   - RawFileFrameInfo - provide access to the elements of a frame header
   - RawFileChannelInfo - provide access to the elements of a channel header
   - RawFileChannelImage - provide access to the channel image data

Exceptions
   - RawFileReaderError - module's exception class

Functions
   - supportedRawFileFormats() - provide a list of the supported raw file formats


Examples

   Below are several examples how the classes in this module can be used to retrieve
   the various parts of the Camera raw files.

   #
   # Obtain raw file header, and last frame, last channel information:
   #
   from camera.formats.raw import *

   f = open( path, "rb" )
   reader = RawFileReader( f )

   # can also use: reader = RawFileReader( path )

   print( reader.header )

   frame = reader.frameCount - 1
   channel = reader.channelCount - 1

   print( reader.frameInfo( frame ) )
   print( reader.channelInfo( frame, channel ) )
   print( reader.channelImage( frame, channel ) )

   #
   # Iterate over all frame information headers:
   #
   for info in reader.frameInfoIter():
      print( 'frame {frm}:'.format( frm=info.frame ) )
      print( info )

   #
   # Iterate over all channel information headers:
   #
   for info in reader.channelInfoIter():
      print( 'channel {frm}.{chn}:'.format( frm=info.frame, chn=info.channel ) )
      print( info )

   #
   # Iterate over all channel images:
   #
   for image in reader.imageIter():
      print( 'image {frm}.{chn}:'.format( frm=image.frame, chn=image.channel ) )
      print( image )
      print()


References

   The C++ code with reading of the raw file and relevant data structures are at:

   method CScanDocument::BuildInternalDataStructureFromRAW()
   - https://svn.eld.leidenuniv.nl/repos/ip-camera/application/trunk/Camera/ScanDocument.h
   - https://svn.eld.leidenuniv.nl/repos/ip-camera/application/trunk/Camera/ScanDocument.cpp
   - see Python class RawFileReader, method _createIndex()

   struct __SpmScanPosition
   - https://svn.eld.leidenuniv.nl/repos/ip-camera/application/trunk/Camera/InterfaceDll.h
   - see Python class RawFileFrameScanArea

   struct __SpmCameraDeviceSpecificFrameData
   - https://svn.eld.leidenuniv.nl/repos/ip-camera/application/trunk/CameraDll/CameraDll.h
   - see Python class RawFileFrameDeviceInfo


Weblinks

   IP: Interface Physics, http://www.physics.leidenuniv.nl/sections/cm/ip/
   LPM: Leiden Probe Microscopy, http://www.leidenprobemicroscopy.com/

   See also http://www.physics.leidenuniv.nl/sections/cm/ip/Internal/Camera/Camera-MainPage.html

"""

__all__ = [
   "supportedRawFileFormats",
   "RawFileReaderError",
   "RawFileReader",
   "RawFileHeader",
   "RawFileFrameInfo",
   "RawFileChannelInfo",
   "RawFileChannelImage", ]

__version_date__ = '$Date: 2012-02-27 14:41:59 +0100 (Mon, 27 Feb 2012) $'
__version__      = '$Revision: 649 $'
__author__       = 'Martin Moene <m.j.moene@eld.physics.LeidenUniv.nl>'
__url__          = 'http://www.physics.leidenuniv.nl/sections/cm/ip/Internal/Camera/Camera-MainPage.html'

import io         # Python 2.6 and 3: io.IoBase file type
import string
import sys
import time

try:
   import numpy
except Exception as detail:
   print( detail )
   print( "Please install the NumPy package (http://numpy.scipy.org/)")
   sys.exit(1)

from . import util

# enforce new-style classes in Python 2.x
if sys.version_info[0] < 3:
   class _metacls(type):
      pass
   __metaclass__ = _metacls

##
# Configuration
#
cfgSupportedRawFileFormats = [ 'RAW1305' ]

cfgTimeFormat = '%Y-%m-%d %H:%M:%S'  # 2010-06-08 10:18:31

##
# keys for class RawFileHeader data dictionary
#
_strFileVersionId = 'fileVersionId'
_strInterfaceId = 'interfaceId'
_strInterfaceVersionMajor = 'interfaceVersionMajor'
_strInterfaceVersionMinor = 'interfaceVersionMinor'
_strFrameCount = 'frameCount'
_strChannelMask = 'channelMask'
_strChannelXName = 'channel{0}Name'
_strChannelXUnit = 'channel{0}Unit'


def supportedRawFileFormats():
   """Provide a list of raw file formats supported by this module, e.g. ['RAW1305']."""
   return cfgSupportedRawFileFormats


class RawFileReaderError( Exception ):

   """Exception class for this module."""

   def __init__( self, value ):
      """Create the exception with the given value."""
      self.value = value
   def __str__( self ):
      """String representation."""
      return repr( self.value )


class RawFileReader:

   """Reader for Camera raw files."""

   def __init__(self, inp):
      """Create the reader from a file object or from a path.

      Keyword arguments:
      inp -- the stream (file) object, or the path to the file
      """
      self._fh = None
      self._filename = None
      if isinstance(inp, io.IOBase):
         self._fh = inp
      elif isinstance(inp, basestring):
         self._filename = inp
         self._fh = open( inp, mode="rb" )
      else:
         raise RawFileReaderError('RawFileReader: ' \
            'invalid constructor argument type {typ}'.format(typ=type(inp)) )

      self._header = RawFileHeader(self._fh)
      self._timeformat = cfgTimeFormat
      self._index = None
      self._createIndex(self._fh)

   def __del__(self):
      """Clean-up: close file if we opened it."""
      if not self._filename is None and not self._fh is None:
         self._fh.close()

   @property
   def timeformat(self):
      """'timeformat' property. See also strftime(),
      http://opengroup.org/onlinepubs/007908799/xsh/strftime.html"""
      return self._timeformat

   @timeformat.setter
   def timeformat(self, fmt):
      """'timeformat' property."""
      self._timeformat = fmt

   @property
   def header(self):
      """'header' property. See also RawFileHeader."""
      return self._header

   @header.setter
   def header(self, value):
      # Prevent assignment to header.
      pass

   @property
   def frameCount(self):
      """'frameCount' property."""
      return self.header.frameCount

   @property
   def channelCount(self):
      """'channelCount' property."""
      return self.header.channelCount

   def frameInfo(self, fi):
      """Provide the frame information of the frame given by its index.
      Note that the frames are selected by their sequence number in the file,
      not by their frameId; these may differ if frames have been deleted.

      Keyword arguments:
      fi -- frame index

      See also RawFileFrameInfo.
      """
      finfo = RawFileFrameInfo(self._fh, self._frameToInfoPos(fi), fi )
      finfo.timeformat = self._timeformat
      return finfo

   def channelInfo(self, fi, ci):
      """Provide the channel information of the channel given by its indices.
      Note that the frames are selected by their sequence number in the file,
      not by their frameId; these may differ if frames have been deleted.

      Keyword arguments:
      fi -- frame index
      ci -- channel index

      See also RawFileChannelInfo.
      """
      return RawFileChannelInfo(self._fh, self._channelToInfoPos(fi, ci), fi, ci )

   def channelImage(self, fi, ci):
      """Provide the image of the channel given by its indices.
      Note that the frames are selected by their sequence number in the file,
      not by their frameId; these may differ if frames have been deleted.

      Keyword arguments:
      fi -- frame index
      ci -- channel index

      See also RawFileChannelImage.
      """
      # deterine image size:
      cinfo = self.channelInfo( fi, ci )
      return RawFileChannelImage(self._fh, self._channelToImagePos(fi, ci), cinfo )

   def frameInfoIter(self):
      """Iterator for frame information, see also RawFileFrameInfo."""
      for index in range( self.frameCount ):
         yield self.frameInfo(index)

   def channelInfoIter(self):
      """Iterator for channel information, see also RawFileChannelInfo."""
      for fi in range( self.frameCount ):
         for ci in range( self.channelCount ):
            yield self.channelInfo(fi, ci)

   def imageIter(self):
      """Iterator for images, see also RawFileChannelImage."""
      for fi in range( self.frameCount ):
         for ci in range( self.channelCount ):
            yield self.channelImage( fi, ci )

   def _createIndex(self, f):
      """Create frame, channel number, info, image to file position index."""
      self._index = {}

      pos = self.header.sizeOf
      for fi in range( self.frameCount ):
         self._saveFrameInfoPos( fi, pos )
         finfo = RawFileFrameInfo( self._fh, pos )

         csize = 0
         for ci in range( self.channelCount ):
            self._saveChannelInfoPos( fi, ci, pos + finfo.sizeOf + csize )
            cinfo  = RawFileChannelInfo( self._fh, pos + finfo.sizeOf + csize )
            csize += cinfo.sizeOf
            self._saveChannelImagePos( fi, ci, pos + finfo.sizeOf + csize )
            csize += cinfo.sizeOfImage

         pos += finfo.sizeOf + csize

   def _saveFrameInfoPos( self, fi, pos ):
      """Save frame information file position."""
      self._validateFrameIndex( fi )
      self._index[ _makeFrameInfoIndexName(fi) ] = pos

   def _saveChannelInfoPos( self, fi, ci, pos ):
      """Save channel information file position."""
      self._validateChannelIndex( fi, ci )
      self._index[ _makeChannelInfoIndexName(fi, ci) ] = pos

   def _saveChannelImagePos( self, fi, ci, pos ):
      """Save channel image file position."""
      self._validateChannelIndex( fi, ci )
      self._index[ _makeChannelImageIndexName(fi, ci) ] = pos

   def _frameToInfoPos( self, fi ):
      """The file position for the information of the given frame."""
      self._validateFrameIndex( fi )
      return self._index[ _makeFrameInfoIndexName(fi) ]

   def _channelToInfoPos( self, fi, ci ):
      """The file position for the information of the given channel."""
      self._validateChannelIndex( fi, ci )
      return self._index[ _makeChannelInfoIndexName(fi, ci) ]

   def _channelToImagePos( self, fi, ci ):
      """The file position for the image of the given channel."""
      self._validateChannelIndex( fi, ci )
      return self._index[ _makeChannelImageIndexName(fi, ci) ]

   def _validateFrameIndex( self, fi ):
      """Validate the given frame index; raise excpetion if the index is not valid."""
      nmax = self.frameCount - 1
      if not ( 0 <= fi and fi <= nmax ):
         raise RawFileReaderError( 'RawFileReader: ' \
            'invalid frame number requested {fi} [0..{nmax}]'.format(fi=fi, nmax=nmax))

   def _validateChannelIndex( self, fi, ci ):
      """Validate the given channel index; raise excpetion if the index is not valid."""
      self._validateFrameIndex( fi )
      nmax = self.channelCount - 1
      if not ( 0 <= ci and ci <= nmax ):
         raise RawFileReaderError( 'RawFileReader: ' \
            'invalid channel number requested {ci} [0..{nmax}]'.format(ci=ci, nmax=nmax))


def _makeFrameInfoIndexName( fi ):
   """The index name for the given frame info index: fi.i (info)"""
   return '{fi:d}.i'.format( fi=fi )

def _makeChannelInfoIndexName( fi, ci ):
   """The index name for the given channel info index: fi.ci.i (info)"""
   return '{fi:d}.{ci:d}.i'.format( fi=fi, ci=ci )

def _makeChannelImageIndexName( fi, ci ):
   """The index name for the given channel image index: fi.ci.g (image, 'graphic')"""
   return '{fi:d}.{ci:d}.g'.format( fi=fi, ci=ci )


class RawFileHeader:

   """
   Class to read and represent a header of a Camera raw file.

   The header class can be treated as a dictionary and the funtions items(),
   keys() and values() enable iteration over it.

   Please note that this (currently) is not true for the classes RawFileFrameInfo
   and RawFileChannelInfo. It is not yet clear to me (Martin) if being a dictionary
   is a useful property of the class.

   """

   def __init__(self):
      """Default constructor"""
      self._db = dict()
      self._db[_strFileVersionId] = None
      self._sizeof = None

   def __init__(self, f):
      """Create the header from a file object."""
      self._db = dict()
      self._read(f)

   def __iter__(self):
      """Provide iterator on header"""
      return self._db.__iter__()

   def __getitem__(self, k):
      """Provide item"""
      return self._db[k]

   def items(self):
      """Provide iteration over items"""
      return self._db.items()

   def keys(self):
      """Provide iteration over keys"""
      return self._db.keys()

   def values(self):
      """Provide iteration over values"""
      return self._db.values()

   def __str__(self):
      """String representation."""
      s = 'header: file version id: {0}\n'.format(self.fileVersionId) + \
          'header: interface id: {0}\n'.format(self.interfaceId) + \
          'header: interface version: {major}.{minor}\n'.format(major=self.interfaceVersionMajor, minor=self.interfaceVersionMinor) + \
          'header: frame count: {0}\n'.format(self.frameCount) + \
          'header: channel mask: 0x{0:x}, 0b{0:b}\n'.format(self.channelMask)  + \
          'header: channel count: {0}\n'.format(self.channelCount)
      for x in range(0,8):
         s += 'header: channel{0} name: {1}\n'.format(x, self.channelName(x))
      for x in range(0,8):
         s += 'header: channel{0} unit: {1}\n'.format(x, self.channelUnit(x))
      return s

   def _read(self, f):
      """Read the header information; raise exception if this file format cannot be handled."""
      f.seek(0)
      self._db[_strFileVersionId] = util.readZString(f)
      self._checkFileVersionId( self.fileVersionId )

      self._db[_strInterfaceId] = util.readZString(f)
      self._db[_strInterfaceVersionMajor] = util.readInt16(f)
      self._db[_strInterfaceVersionMinor] = util.readInt16(f)
      self._db[_strFrameCount] = util.readInt32(f)
      self._db[_strChannelMask] = util.readInt32(f)
      for i in range(0,8):
         self._db[_strChannelXName.format(i)] = util.readZString(f)
      for i in range(0,8):
         self._db[_strChannelXUnit.format(i)] = util.readZString(f)
      self._sizeof = f.tell()

   def _checkFileVersionId(self, versionId):
      """check if this raw file version can be processed
      - RAW1305:
      """
      if 'RAW' != versionId[0:3]:
         raise RawFileReaderError('this is probably not a Camera raw file')

      versionNum = int(versionId[3:])
      for x in cfgSupportedRawFileFormats:
         validNum = int(x[3:])
#         print( 'version id:{0} num:{1}'.format(versionId, versionNum) )
         if versionId == x:
            return
         elif versionNum < validNum:
            raise RawFileReaderError('cannot handle raw file with version number older than ' \
               '{valid}: {vnum} ({vid})'.format(vnum=versionNum, vid=versionId, valid=validNum) )
         elif versionNum > validNum:
            raise RawFileReaderError('cannot handle raw file with version number newer than ' \
               '{valid}: {vnum} ({vid})'.format(vnum=versionNum, vid=versionId, valid=validNum) )

   @property
   def sizeOf(self):
      """'sizeOf' property."""
      return self._sizeof

   @property
   def fileVersionId(self):
      """'versionId' property."""
      return self._db[_strFileVersionId]

   @fileVersionId.setter
   def fileVersionId(self, value):
      self._db[_strFileVersionId] = value

   @property
   def interfaceId(self):
      """'interfaceId' property."""
      return self._db[_strInterfaceId]

   @interfaceId.setter
   def interfaceId(self, value):
      self._db[_strInterfaceId] = value

   @property
   def interfaceVersionMajor(self):
      """'interfaceVersionMajor' property."""
      return int(self._db[_strInterfaceVersionMajor])

   @interfaceVersionMajor.setter
   def interfaceVersionMajor(self, value):
      self._db[_strInterfaceVersionMajor] = int(value)

   @property
   def interfaceVersionMinor(self):
      """'interfaceVersionMinor' property."""
      return int(self._db[_strInterfaceVersionMinor])

   @interfaceVersionMinor.setter
   def interfaceVersionMinor(self, value):
      self._db[_strInterfaceVersionMinor] = int(value)

   @property
   def frameCount(self):
      """'frameCount' property."""
      return int(self._db[_strFrameCount])

   @frameCount.setter
   def frameCount(self, value):
      self._db[_strFrameCount] = int(value)

   @property
   def channelMask(self):
      """'channelMask' property."""
      return int(self._db[_strChannelMask])

   @channelMask.setter
   def channelMask(self, value):
      self._db[_strChannelMask] = int(value)

   @property
   def channelCount(self):
      """'channelCount' derived (ro) property."""
      return util.bitcount( self.channelMask )

   # channel names

   def channelName(self, i):
      """Name of the specified channel."""
      return self._db[_strChannelXName.format(i)]

   def channelUnit(self, i):
      """Unit of the specified channel."""
      return self._db[_strChannelXUnit.format(i)]

   @property
   def channel0Name(self):
      """'channel0Name' property."""
      return self._db[_strChannelXName.format(0)]

   @channel0Name.setter
   def channel0Name(self, value):
      self._db[_strChannelXName.format(0)] = value

   @property
   def channel1Name(self):
      """'channel1Name' property."""
      return self._db[_strChannelXName.format(1)]

   @channel1Name.setter
   def channel1Name(self, value):
      self._db[_strChannelXName.format(1)] = value

   @property
   def channel2Name(self):
      """'channel2Name' property."""
      return self._db[_strChannelXName.format(2)]

   @channel2Name.setter
   def channel2Name(self, value):
      self._db[_strChannelXName.format(2)] = value

   @property
   def channel3Name(self):
      """'channel0Name' property."""
      return self._db[_strChannelXName.format(3)]

   @channel3Name.setter
   def channel3Name(self, value):
      self._db[_strChannelXName.format(3)] = value

   @property
   def channel4Name(self):
      """'channel4Name' property."""
      return self._db[_strChannelXName.format(4)]

   @channel4Name.setter
   def channel4Name(self, value):
      self._db[_strChannelXName.format(4)] = value

   @property
   def channel5Name(self):
      """'channel5Name' property."""
      return self._db[_strChannelXName.format(5)]

   @channel5Name.setter
   def channel5Name(self, value):
      self._db[_strChannelXName.format(5)] = value

   @property
   def channel6Name(self):
      """'channel6Name' property."""
      return self._db[_strChannelXName.format(6)]

   @channel6Name.setter
   def channel6Name(self, value):
      self._db[_strChannelXName.format(6)] = value

   @property
   def channel7Name(self):
      """'channel0Name' property."""
      return self._db[_strChannelXName.format(7)]

   @channel7Name.setter
   def channel0Name(self, value):
      self._db[_strChannelXName.format(7)] = value

   # channel units

   @property
   def channel0Unit(self):
      """'channel0Unit' property."""
      return self._db[_strChannelXUnit.format(0)]

   @channel0Unit.setter
   def channel0Unit(self, value):
      self._db[_strChannelXUnit.format(0)] = value

   @property
   def channel1Unit(self):
      """'channel1Unit' property."""
      return self._db[_strChannelXUnit.format(1)]

   @channel1Unit.setter
   def channel1Unit(self, value):
      self._db[_strChannelXUnit.format(1)] = value

   @property
   def channel2Unit(self):
      """'channel2Unit' property."""
      return self._db[_strChannelXUnit.format(2)]

   @channel2Unit.setter
   def channel2Unit(self, value):
      self._db[_strChannelXUnit.format(2)] = value

   @property
   def channel3Unit(self):
      """'channel0Unit' property."""
      return self._db[_strChannelXUnit.format(3)]

   @channel3Unit.setter
   def channel3Unit(self, value):
      self._db[_strChannelXUnit.format(3)] = value

   @property
   def channel4Unit(self):
      """'channel4Unit' property."""
      return self._db[_strChannelXUnit.format(4)]

   @channel4Unit.setter
   def channel4Unit(self, value):
      self._db[_strChannelXUnit.format(4)] = value

   @property
   def channel5Unit(self):
      """'channel5Unit' property."""
      return self._db[_strChannelXUnit.format(5)]

   @channel5Unit.setter
   def channel5Unit(self, value):
      self._db[_strChannelXUnit.format(5)] = value

   @property
   def channel6Unit(self):
      """'channel6Unit' property."""
      return self._db[_strChannelXUnit.format(6)]

   @channel6Unit.setter
   def channel6Unit(self, value):
      self._db[_strChannelXUnit.format(6)] = value

   @property
   def channel7Unit(self):
      """'channel0Unit' property."""
      return self._db[_strChannelXUnit.format(7)]

   @channel7Unit.setter
   def channel0Unit(self, value):
      self._db[_strChannelXUnit.format(7)] = value


class RawFileFrameInfo:

   """Class to read and represent the header of a frame of a Camera raw file."""

   def __init__(self, f, pos, frame=None):
      """Create the collection from a file object at the specified position.
      Note that the frames are selected by their sequence number in the file,
      not by their frameId; index and id may differ if frames have been deleted.

      Keyword arguments:
      f -- stream object
      pos -- position in stream
      frame -- frame index
      """
      self._frame = frame
      self._timeformat = cfgTimeFormat
      self._read(f, pos)
      pass

   def _read(self, f, pos):
      """Read the frame information from the specified file position."""
#      logger.debug( 'frame info f.seek({0})'.format( pos ) )
      f.seek( pos )
      self._frameId = util.readInt32(f)
      self._acquisitionTime = util.readInt32(f)

      # ScanPosition struct (area) :
      self._scanArea = RawFileFrameScanArea(f, self.frame)

      # device-specific data struct (device):
      self._deviceInfo = RawFileFrameDeviceInfo(f, self.frame)

   def __str__(self):
      """String representation."""
      return 'frame {frm}: id: {fid}\n'.format( frm=self.frame, fid=self.frameId ) + \
             'frame {frm}: acquisition time: {t1}, {t2}\n'.format( frm=self.frame, t1=self.acquisitionTime, t2=time.strftime(self._timeformat, time.localtime( self.acquisitionTime) ) ) + \
             str( self._scanArea ) + \
             str( self._deviceInfo )

   @property
   def frame(self):
      """'frame' property."""
      return self._frame

   @property
   def timeformat(self):
      """'timeformat' property."""
      return self._timeformat

   @timeformat.setter
   def timeformat(self, fmt):
      """'frame' property."""
      self._timeformat = fmt

   @property
   def sizeOf(self):
      """'sizeOf' property."""
      return 4 + 4 + self.sizeOfScanArea + self.sizeofDeviceInfo

   @property
   def frameId(self):
      """'frameId' property."""
      return self._frameId

   @property
   def acquisitionTime(self):
      """'acquisitionTime' property; seconds since epoch (1970)."""
      return self._acquisitionTime

   @property
   def scanArea(self):
      """'scanArea' property."""
      return self._scanArea

   @property
   def deviceInfo(self):
      """'deviceInfo' property."""
      return self._deviceInfo

   # RawFileFrameScanArea:

   @property
   def sizeOfScanArea(self):
      """'sizeofInfo' property."""
      return self.scanArea.sizeOf

   @property
   def xoffset_pm(self):
      """'x-offset [pm]' property."""
      return self.scanArea.xoffset_pm

   @property
   def yoffset_pm(self):
      """'y-offset [pm]' property."""
      return self.scanArea.yoffset_pm

   @property
   def xstep_pm(self):
      """'x-step [pm]' property."""
      return self.scanArea.xstep_pm

   @property
   def ystep_pm(self):
      """'y-step [pm]' property."""
      return self.scanArea.ystep_pm

   @property
   def xsize_pix(self):
      """'x-size [pixels]' property."""
      return self.scanArea.xsize_pix

   @property
   def ysize_pix(self):
      """'y-size [pixels]' property."""
      return self.scanArea.ysize_pix

   @property
   def scanangle_rad(self):
      """'scan angle [rad]' property."""
      return self.scanArea.scanangle_rad

   # RawFileFrameDeviceInfo:

   @property
   def sizeofDeviceInfo(self):
      """'sizeofInfo' property."""
      return self.deviceInfo._sizeofInfo

   @property
   def pixelclock_kHz(self):
      """'pixelclock_kHz' property."""
      return self.deviceInfo._pixelclock_kHz

   @property
   def samplesPerPoint(self):
      """'samplesPerPoint' property."""
      return self.deviceInfo._samplesPerPoint

   @property
   def oversamplingFactor(self):
      """'oversamplingFactor' property."""
      return self.deviceInfo._oversamplingFactor

   @property
   def zgain(self):
      """'zgain' property."""
      return self.deviceInfo._zgain

   @property
   def dzgain(self):
      """'dzgain' property."""
      return self.deviceInfo._dzgain

   @property
   def bandwidth_kHz(self):
      """'bandwidth_kHz' property."""
      return self.deviceInfo._bandwidth_kHz

   @property
   def zoffset_QQ(self):
      """'zoffset_QQ' property."""
      return self.deviceInfo._zoffset_QQ

   @property
   def zoverdz(self):
      """'zoverdz' property."""
      return self.deviceInfo._zoverdz

   @property
   def xrounding_pct(self):
      """'xrounding_pct' property."""
      return self.deviceInfo._xrounding_pct

   @property
   def backwardstroke(self):
      """'backwardstroke' property."""
      return self.deviceInfo._backwardstroke

   @property
   def usample_V(self):
      """'usample_V' property."""
      return self.deviceInfo._usample_V

   @property
   def triangleAttenuation(self):
      """'triangleAttenuation' property."""
      return self.deviceInfo._triangleAttenuation

   @property
   def sawtoothAttenuation(self):
      """'sawtoothAttenuation' property."""
      return self.deviceInfo._sawtoothAttenuation

   @property
   def xsampleSlope(self):
      """'xsampleSlope' property."""
      return self.deviceInfo._xsampleSlope

   @property
   def ysampleSlope(self):
      """'ysampleSlope' property."""
      return self.deviceInfo._ysampleSlope

   @property
   def integrationTimeCode(self):
      """'integrationTimeCode' property."""
      return self.deviceInfo._integrationTimeCode

   @property
   def proportionalGainCode(self):
      """'proportionalGainCode' property."""
      return self.deviceInfo._proportionalGainCode

   @property
   def setvalueCode(self):
      """'setvalueCode' property."""
      return self.deviceInfo._setvalueCode

   @property
   def d3MeasurementsPerPixel(self):
      """'3dMeasurementsPerPixel' property."""
      return self.deviceInfo._3dMeasurementsPerPixel

   @property
   def d3DisplayIndex(self):
      """'3dDisplayIndex' property."""
      return self.deviceInfo._3dDisplayIndex

   @property
   def d3FrameNumber(self):
      """'3dFrameNumber' property."""
      return self.deviceInfo._3dFrameNumber

   @property
   def extraMixerOffset(self):
      """'extraMixerOffset' property."""
      return self.deviceInfo._extraMixerOffset

   @property
   def extraMixerXgain(self):
      """'extraMixerXgain' property."""
      return self.deviceInfo._extraMixerXgain

   @property
   def extraMixerYgain(self):
      """'extraMixerYgain' property."""
      return self.deviceInfo._extraMixerYgain

   @property
   def extraMixerAttenuation(self):
      """'extraMixerAttenuation' property."""
      return self.deviceInfo._extraMixerAttenuation

   @property
   def extraMixerSlewrate(self):
      """'extraMixerSlewrate' property."""
      return self.deviceInfo._extraMixerSlewrate

   @property
   def adcChannels(self):
      """'adcChannels' property; yields list."""
      return self.deviceInfo._adcChannels

   @property
   def ivname(self):
      """'ivname' property."""
      return self.deviceInfo._ivname


#
# RAW1305
#
#typedef struct __SpmScanPosition
#{
#	int			nXOffset;		// XOffset of center of area in picometer
#	int			nYOffset;		// YOffset of center of area in picometer
#	double		dXStepSize;		// X Step size of area in picometer
#	double		dYStepSize;		// Y Step size of area in picometer
#	DWORD		dwOrgScanXSize;	// Original X Size in pixels
#	DWORD		dwOrgScanYSize;	// Original X Size in pixels
#	double		dScanAngle;		// Rotation of scan around center in rad.
#} SPMSCANPOSITION, *LPSPMSCANPOSITION;
#

class RawFileFrameScanArea:

   """Class to read and represent a measurement area (SpmScanPosition) of a
   Camera raw file.
   """

   def __init__(self, f, frame):
      """Create the scan area from a file object at the current position.

      Keyword arguments:
      f -- stream object
      frame -- frame index

      See also RawFileFrameInfo.
      """
      self._frame = frame
      self._read(f)
      pass

   def _read(self, f):
      """Read the frame scan area information from the current file position"""
      self._xoffset_pm = util.readInt32(f)
      self._yoffset_pm = util.readInt32(f)
      self._xstep_pm = util.readDouble(f)
      self._ystep_pm = util.readDouble(f)
      self._xorigscansize_pix = util.readInt32(f)
      self._yorigscansize_pix = util.readInt32(f)
      self._scanangle_rad = util.readDouble(f)

   def __str__(self):
      """String representation."""
      return 'frame {frm}: x-offset [pm]: {x}\n'.format( frm=self.frame, x=self.xoffset_pm ) + \
             'frame {frm}: y-offset [pm]: {x}\n'.format( frm=self.frame, x=self.yoffset_pm ) + \
             'frame {frm}: x-step [pm]: {x}\n'.format( frm=self.frame, x=self.xstep_pm ) + \
             'frame {frm}: x-step [pm]: {x}\n'.format( frm=self.frame, x=self.ystep_pm ) + \
             'frame {frm}: x-size [pixels]: {x}\n'.format( frm=self.frame, x=self.xsize_pix ) + \
             'frame {frm}: y-size [pixels]: {x}\n'.format( frm=self.frame, x=self.ysize_pix ) + \
             'frame {frm}: scan angle [rad]: {x}\n'.format (frm=self.frame, x=self.scanangle_rad )

   @property
   def frame(self):
      """'frame' property."""
      return self._frame

   @property
   def sizeOf(self):
      """'sizeOf' property."""
      return 40  # 2 * 4 + 2 * 8 + 2 * 4 + 8

   @property
   def xoffset_pm(self):
      """'x-offset [pm]' property."""
      return self._xoffset_pm

   @property
   def yoffset_pm(self):
      """'y-offset [pm]' property."""
      return self._yoffset_pm

   @property
   def xstep_pm(self):
      """'x-step [pm]' property."""
      return self._xstep_pm

   @property
   def ystep_pm(self):
      """'y-step [pm]' property."""
      return self._ystep_pm

   @property
   def xsize_pix(self):
      """'x-size [pixels]' property."""
      return self._xorigscansize_pix

   @property
   def ysize_pix(self):
      """'y-size [pixels]' property."""
      return self._yorigscansize_pix

   @property
   def scanangle_rad(self):
      """'scan angle [rad]' property."""
      return self._scanangle_rad


#
# RAW1305
#
#typedef struct __SpmCameraDeviceSpecificFrameData
#{
#	///////////////////////////////////
#	//NB. NOG HETZELFDE ALS PREV1 !!!!!
#	///////////////////////////////////
#
#	DWORD		dwSize;			// Size of this structure in bytes
#	//
#	// device specific frame data
#	//
#	double	dPixelClock;			//pixelclock
#	UINT	uSamplingRate;			//sampling rate (.. samples /point for triangular scan movement)
#	UINT	uOverSampling;			//Oversampling most of the time 8x
#
#	// Peter TODO eigenlijk moet de gain van alle gebruikte kanalen in deze frame komen
#	UINT	uZAmpGain;				//Z Amplifier gain
#	UINT	uDeltaZAmpGain;			//dZ Amplifier gain
#	UINT	uBandWidth;				//Bandwidth
#	double	dZOffset;				//ZOffset
#	double  dZdZFactor;				//Z/dZ factor
#
#	double	dXRounding;				//X Rounding
#	UINT	uBackwardStroke	;		//Backward stroke
#
#	double dSampleVoltage;			//sample volatge
#
#	UINT	uAttenuatorTriangular;	//Attenuator triangular scan movement
#	UINT	uAttenuatorSawTooth;	//Attenuator saw tooth scan movement
#	UINT	uSampleSlopeX;			//Sample slope in x-direction
#	UINT	uSampleSlopeY;			//Sample slope in y-direction
#
#	UINT	uIntegrationTime;		//integration time regulator
#	UINT	uProportionalGain;		//proportional gain regulator
#	UINT	uWishedTunnelingCurrent;//wished tunneling current
#
#	// Peter 3d : hoe geef ik parameters door aan het hoofdprogramma
#	int		i3dMeasPerPixel;		//igv 3d aantal meteingen per pixel, anders 0
#	int		i3dDisplayIndex;		//igv 3d index in array om te displayen
#	int		i3dFrameNumber;			//igv 3d line-number
#
#	// Peter Extra Mixer
#	double	dExtraMixerOffset;		// Extra Mixer Offset in [V]
#	int	iExtraMixerAmpX;			// Extra Mixer Amplification X in bits
#	int iExtraMixerAmpY;			// Extra Mixer Amplification Y in bits
#	UINT uExtraMixerAttenuation;	// Extra Mixer Attenuation 1:gain=1,2:gain=0.1,3:gain=0.01,4=gain=0.001
#	UINT uExtraMixerSlewRate;		// Extra Mixer SlewRate in bits 5..65535
#
#	// Peter multiADC
#	BOOL	bADCchannels[8];		// ADC kanalen in gebruik
#
#	char	csIVName[MAX_PATH];		// Linked I/V name
#} SPMCAMERADEVICESPECIFICFRAMEDATA, *LPSPMCAMERADEVICESPECIFICFRAMEDATA;
#

class RawFileFrameDeviceInfo:

   """Class to read and represent a frame's device information
   (__SpmCameraDeviceSpecificFrameData) of a Camera raw file.
   """

   def __init__(self, f, frame):
      """Create the device information from a file object at the current position.

      Keyword arguments:
      f -- stream object
      frame -- frame index

      See also RawFileFrameInfo.
      """
      self._frame = frame
      self._read(f)
      pass

   def _read(self, f):
      """Read the frame device information from the current file position."""
      self._sizeofInfo = util.readInt32(f)
      padding = util.readInt32(f)

      self._pixelclock_kHz = util.readDouble(f)         # ToDo Hz??
      self._samplesPerPoint = util.readInt32(f)         # samples / point for triangular scan movement
      self._oversamplingFactor = util.readInt32(f)      # mostly 8x

      self._zgain = util.readInt32(f)
      self._dzgain = util.readInt32(f)
      self._bandwidth_kHz = util.readInt32(f)
      padding = util.readInt32(f)
      self._zoffset_QQ = util.readDouble(f)
      self._zoverdz = util.readDouble(f)

      self._xrounding_pct = util.readDouble(f)
      self._backwardstroke = util.readInt32(f)          # meaning?

      padding = util.readInt32(f)
      self._usample_V = util.readDouble(f)

      self._triangleAttenuation = util.readInt32(f)     # attenuator triangular scan movement ??
      self._sawtoothAttenuation = util.readInt32(f)     # attenuator saw tooth scan movement ??
      self._xsampleSlope = util.readInt32(f)
      self._ysampleSlope = util.readInt32(f)

      self._integrationTimeCode = util.readInt32(f)
      self._proportionalGainCode = util.readInt32(f)
      self._setvalueCode = util.readInt32(f)            # desired tunneling current

      self._3dMeasurementsPerPixel = util.readInt32(f)  # for 3d: # measurements per pixel, 0 othersize
      self._3dDisplayIndex = util.readInt32(f)          # for 3d: index in array to display
      self._3dFrameNumber = util.readInt32(f)           # for 3d: line-number

      self._extraMixerOffset = util.readDouble(f)
      self._extraMixerXgain = util.readInt32(f)
      self._extraMixerYgain = util.readInt32(f)
      self._extraMixerAttenuation = util.readInt32(f)   # 1:gain=1,2:gain=0.1,3:gain=0.01,4=gain=0.001
      self._extraMixerSlewrate = util.readInt32(f)      # Extra Mixer SlewRate in bits 5..65535

      self._adcChannels = numpy.fromfile(f, dtype='<i4', count=8)
                                                    # actually filled since Camera 4.0.0

      self._ivname = util.readZString(f)                # Linked I/V name

   def __str__(self):
      """String representation."""
      return 'frame {frm}: sizeof device info: {x}\n'.format( frm=self.frame, x=self.sizeofDeviceInfo) + \
             'frame {frm}: pixel clock [kHz]: {x}\n'.format( frm=self.frame, x=self.pixelclock_kHz) + \
             'frame {frm}: sample per point: {x}\n'.format( frm=self.frame, x=self.samplesPerPoint) + \
             'frame {frm}: oversampling factor: {x}\n'.format( frm=self.frame, x=self.oversamplingFactor) + \
             'frame {frm}: z-gain: {x}\n'.format( frm=self.frame, x=self.zgain) + \
             'frame {frm}: dz-gain: {x}\n'.format( frm=self.frame, x=self.dzgain) + \
             'frame {frm}: bandwidth [kHz]: {x}\n'.format( frm=self.frame, x=self.bandwidth_kHz) + \
             'frame {frm}: z-offset [??]: {x}\n'.format( frm=self.frame, x=self.zoffset_QQ) + \
             'frame {frm}: zoverdz: {x}\n'.format( frm=self.frame, x=self.zoverdz) + \
             'frame {frm}: x-rounding [%]: {x}\n'.format( frm=self.frame, x=self.xrounding_pct) + \
             'frame {frm}: backwardstroke: {x}\n'.format( frm=self.frame, x=self.backwardstroke) + \
             'frame {frm}: u-sample [V]: {x}\n'.format( frm=self.frame, x=self.usample_V) + \
             'frame {frm}: triangle attenuation: {x}, 0x{x:x}\n'.format( frm=self.frame, x=self.triangleAttenuation) + \
             'frame {frm}: sawtooth attenuation: {x}, 0x{x:x}\n'.format( frm=self.frame, x=self.sawtoothAttenuation) + \
             'frame {frm}: x-sample slope: {x}\n'.format( frm=self.frame, x=self.xsampleSlope) + \
             'frame {frm}: y-sample slope: {x}\n'.format( frm=self.frame, x=self.ysampleSlope) + \
             \
             'frame {frm}: integration time code: {x}\n'.format( frm=self.frame, x=self.integrationTimeCode) + \
             'frame {frm}: proportional gain code: {x}\n'.format( frm=self.frame, x=self.proportionalGainCode) + \
             'frame {frm}: setvalue code: {x}\n'.format( frm=self.frame, x=self.setvalueCode) + \
             \
             'frame {frm}: 3D measurements per pixel: {x}\n'.format( frm=self.frame, x=self.d3MeasurementsPerPixel) + \
             'frame {frm}: 3D display index: {x}\n'.format( frm=self.frame, x=self.d3DisplayIndex) + \
             'frame {frm}: 3D frame number: {x}\n'.format( frm=self.frame, x=self.d3FrameNumber) + \
             \
             'frame {frm}: extra mixer offset [V]: {x}\n'.format( frm=self.frame, x=self.extraMixerOffset) + \
             'frame {frm}: extra mixer x-gain: {x}\n'.format( frm=self.frame, x=self.extraMixerXgain) + \
             'frame {frm}: extra mixer y-gain: {x}\n'.format( frm=self.frame, x=self.extraMixerYgain) + \
             'frame {frm}: extra mixer attenuation: {x}\n'.format( frm=self.frame, x=self.extraMixerAttenuation) + \
             'frame {frm}: extra mixer slewrate: {x}\n'.format( frm=self.frame, x=self.extraMixerSlewrate) + \
             \
             'frame {frm}: adc channels: {x}\n'.format( frm=self.frame, x=self.adcChannels) + \
             'frame {frm}: iv name: {x}\n'.format( frm=self.frame, x=self.ivname)

   @property
   def frame(self):
      """'frame' property."""
      return self._frame

   @property
   def sizeofDeviceInfo(self):
      """'sizeofInfo' property."""
      return self._sizeofInfo

   @property
   def pixelclock_kHz(self):
      """'pixelclock_kHz' property."""
      return self._pixelclock_kHz

   @property
   def samplesPerPoint(self):
      """'samplesPerPoint' property."""
      return self._samplesPerPoint

   @property
   def oversamplingFactor(self):
      """'oversamplingFactor' property."""
      return self._oversamplingFactor

   @property
   def zgain(self):
      """'zgain' property."""
      return self._zgain

   @property
   def dzgain(self):
      """'dzgain' property."""
      return self._dzgain

   @property
   def bandwidth_kHz(self):
      """'bandwidth_kHz' property."""
      return self._bandwidth_kHz

   @property
   def zoffset_QQ(self):
      """'zoffset_QQ' property."""
      return self._zoffset_QQ

   @property
   def zoverdz(self):
      """'zoverdz' property."""
      return self._zoverdz

   @property
   def xrounding_pct(self):
      """'xrounding_pct' property."""
      return self._xrounding_pct

   @property
   def backwardstroke(self):
      """'backwardstroke' property."""
      return self._backwardstroke

   @property
   def usample_V(self):
      """'usample_V' property."""
      return self._usample_V

   @property
   def triangleAttenuation(self):
      """'triangleAttenuation' property."""
      return self._triangleAttenuation

   @property
   def sawtoothAttenuation(self):
      """'sawtoothAttenuation' property."""
      return self._sawtoothAttenuation

   @property
   def xsampleSlope(self):
      """'xsampleSlope' property."""
      return self._xsampleSlope

   @property
   def ysampleSlope(self):
      """'ysampleSlope' property."""
      return self._ysampleSlope

   @property
   def integrationTimeCode(self):
      """'integrationTimeCode' property."""
      return self._integrationTimeCode

   @property
   def proportionalGainCode(self):
      """'proportionalGainCode' property."""
      return self._proportionalGainCode

   @property
   def setvalueCode(self):
      """'setvalueCode' property."""
      return self._setvalueCode

   @property
   def d3MeasurementsPerPixel(self):
      """'3dMeasurementsPerPixel' property."""
      return self._3dMeasurementsPerPixel

   @property
   def d3DisplayIndex(self):
      """'3dDisplayIndex' property."""
      return self._3dDisplayIndex

   @property
   def d3FrameNumber(self):
      """'3dFrameNumber' property."""
      return self._3dFrameNumber

   @property
   def extraMixerOffset(self):
      """'extraMixerOffset' property."""
      return self._extraMixerOffset

   @property
   def extraMixerXgain(self):
      """'extraMixerXgain' property."""
      return self._extraMixerXgain

   @property
   def extraMixerYgain(self):
      """'extraMixerYgain' property."""
      return self._extraMixerYgain

   @property
   def extraMixerAttenuation(self):
      """'extraMixerAttenuation' property."""
      return self._extraMixerAttenuation

   @property
   def extraMixerSlewrate(self):
      """'extraMixerSlewrate' property."""
      return self._extraMixerSlewrate

   @property
   def adcChannels(self):
      """'adcChannels' property; yields list."""
      return self._adcChannels

   @property
   def ivname(self):
      """'ivname' property."""
      return self._ivname


class RawFileChannelInfo:

   """Class to read and represent channel information of a Camera raw file."""

   LR = 0x01
   RL = 0x02
   LRRL = LR | RL

   def __init__(self, f, pos, frame=None, channel=None):
      """Create the channel information from a file object at the specified position.

      Keyword arguments:
      f -- stream object
      pos -- position in stream
      frame -- frame index
      channel -- channel index

      See also RawFileFrameInfo.
      """
      self._frame = frame
      self._channel = channel
      self._read(f, pos)
      pass

   def _read(self, f, pos):
      """Read the channel information information from the specified file position."""
#      logger.debug( 'channel info: f.seek({0})'.format( pos ) )
      f.seek( pos )
      self._flags = util.readInt32(f)
      self._xsize = util.readInt32(f)
      self._ysize = util.readInt32(f)
      self._pixels = util.readInt32(f)
      self._dataScaleFactor = util.readDouble(f)
      self._dataOffset = util.readInt32(f)
      self._parameterMask = util.readInt32(f)
      self._xslope = util.readDouble(f)
      self._yslope = util.readDouble(f)
      self._average = util.readDouble(f)    # of what?
      self._zmin = util.readInt32(f)
      self._zmax = util.readInt32(f)

   def __str__(self):
      """String representation."""
      return 'channel {frm}.{chn}: flags: {x}, 0b{x:b}\n'.format( frm=self.frame, chn=self.channel, x=self._flags ) + \
             'channel {frm}.{chn}: x-size: {x}\n'.format( frm=self.frame, chn=self.channel, x=self._xsize ) + \
             'channel {frm}.{chn}: y-size: {x}\n'.format( frm=self.frame, chn=self.channel, x=self._ysize ) + \
             'channel {frm}.{chn}: pixels: {x}\n'.format( frm=self.frame, chn=self.channel, x=self._pixels ) + \
             'channel {frm}.{chn}: data scale factor: {x}\n'.format( frm=self.frame, chn=self.channel, x=self._dataScaleFactor ) + \
             'channel {frm}.{chn}: data offset: {x}\n'.format( frm=self.frame, chn=self.channel, x=self._dataOffset ) + \
             'channel {frm}.{chn}: parameter mask: {x}, 0b{x:b}\n'.format( frm=self.frame, chn=self.channel, x=self._parameterMask ) + \
             'channel {frm}.{chn}: x-slope [?]: {x}\n'.format( frm=self.frame, chn=self.channel, x=self._xslope ) + \
             'channel {frm}.{chn}: y-slope [?]: {x}\n'.format( frm=self.frame, chn=self.channel, x=self._yslope ) + \
             'channel {frm}.{chn}: average [?]: {x}\n'.format( frm=self.frame, chn=self.channel, x=self._average ) + \
             'channel {frm}.{chn}: z-min: {x}\n'.format( frm=self.frame, chn=self.channel, x=self._zmin ) + \
             'channel {frm}.{chn}: z-max: {x}\n'.format( frm=self.frame, chn=self.channel, x=self._zmax )

   @property
   def frame(self):
      """'frame' property."""
      return self._frame

   @property
   def channel(self):
      """'channel' property."""
      return self._channel

   @property
   def sizeOf(self):
      """'sizeOf' property."""
      return int(64)  # 4*4 + 1*8 + 2*4 + 3*8 + 2*4

   @property
   def sizeOfImage(self):
      """'sizeOfImage' property: size (in bytes) of all stored pixels"""
      return self.pixels * 2

   @property
   def flags(self):
      """'flags' property."""
      return self._flags

   @property
   def xsize(self):
      """'xsize' property."""
      return self._xsize

   @property
   def ysize(self):
      """'ysize' property."""
      return self._ysize

   @property
   def pixels(self):
      """'pixels' property."""
      return self._pixels

   @property
   def dataScaleFactor(self):
      """'dataScaleFactor' property."""
      return self._dataScaleFactor

   @property
   def dataOffset(self):
      """'dataOffset' property."""
      return self._dataOffset

   @property
   def parameterMask(self):
      """'parameterMask' property."""
      return self._parameterMask

   @property
   def xslope(self):
      """'xslope' property."""
      return self._xslope

   @property
   def yslope(self):
      """'yslope' property."""
      return self._yslope

   @property
   def average(self):
      """'average' property."""
      return self._average

   @property
   def zmin(self):
      """'zmin' property."""
      return self._zmin

   @property
   def zmax(self):
      """'zmax' property."""
      return self._zmax


class RawFileChannelImage:

   """Class to read and represent channel image data of a Camera raw file."""

   def __init__(self, f, pos, cinfo ):
      """Create the image from the given channel information and from the
      data in the file at the specified position.

      Keyword arguments:
      pos -- position in stream
      cinfo -- channel information object
      """
#      logger.debug( 'channel image: f.seek({0})'.format( pos ) )
      self._frame = cinfo.frame
      self._channel = cinfo.channel
      self._flags = cinfo.flags
      self._xsize = cinfo.xsize
      self._ysize = cinfo.ysize
      self._pixels = cinfo.pixels
      self._read(f, pos)

   def __str__(self):
      """String representation: present an image as 2-dimensional textual data"""
      s = 'image {frm}.{chn}:\n'.format( frm=self.frame, chn=self.channel )
      s += str(self.asArray())
      return s

   def _read(self, f, pos):
      """Read the image pixels from the file at the specified position."""
#      logger.debug( 'channel info: f.seek({0})'.format( pos ) )
      f.seek( pos )
      self._data = numpy.fromfile( f, dtype='<h2', count=self.pixels )

      # self._image will always contain data
      # if set, self._image_LR and self._image_RL have the LR or RL data respectively
      if (self.flags & RawFileChannelInfo.LRRL) == RawFileChannelInfo.LRRL:
         self._image_LR = self._data[:self.pixels // 2]
         self._image_RL = self._data[self.pixels // 2:]

         # reshape only works if the number of data points doesn't change
         # resize only works if there is only one reference to the data, but we now have two
         if self.pixels == self.xsize * self.ysize * 2:
            self._image_LR = self._image_LR.reshape(( self.ysize, self.xsize ))
            self._image_RL = self._image_RL.reshape(( self.ysize, self.xsize ))
         else:
            self._image_LR = self._image_LR.copy()
            self._image_RL = self._image_RL.copy()
            self._image_LR.resize(( self.ysize, self.xsize ))
            self._image_RL.resize(( self.ysize, self.xsize ))

         self._image = self._image_LR
      else:
         self._image = self._data
         self._image.resize(( self._ysize, self._xsize ))
         if self.flags & RawFileChannelInfo.LR:
            self._image_LR = self._image
         else:
            self._image_RL = self._image

   @property
   def frame(self):
      """'frame' property."""
      return self._frame

   @property
   def channel(self):
      """'channel' property."""
      return self._channel

   @property
   def flags(self):
      """'flags' property."""
      return self._flags

   @property
   def xsize(self):
      """'xsize' property."""
      return self._xsize

   @property
   def ysize(self):
      """'ysize' property."""
      return self._ysize

   @property
   def pixels(self):
      """'pixels' property."""
      return self._pixels

   def asArray(self, direction=None):
      """Provide the image data as a two-dimensional array (numpy module)"""
      if direction == RawFileChannelInfo.LR:
         return self._image_LR
      elif direction == RawFileChannelInfo.RL:
         return self._image_RL
      elif direction is None:
         return self._image
      else:
         raise RawFileReaderError( "RawFileChannelImage: direction '{0!r}' not valid".format(direction) )


# ---------------------------------------------------------------------------

def main():
   """Test program."""
   path = sys.argv[1]
   logger.info( "{0}:".format(path))
   f = open(path, "rb")
   print('{0}:'.format(path))

   reader = RawFileReader(f)
   header = reader.header

   print( str(header) )

#   for k in header.keys():
#      print(k, ' ')
#
#   for v in header.values():
#      print(v, ' ')
#
#   for k,v in header.items():
#      print(k, v, ' ')

   info = reader.frameInfo(0)
   print( str(info) )

   info = reader.channelInfo(0, 0)
   print( str(info) )

   info = reader.channelInfo(0, reader.channelCount - 1)
   print( str(info) )

   frame = reader.frameCount - 1
   info = reader.frameInfo(frame)
   print( str(info) )

   info = reader.channelInfo(frame, 0)
   print( str(info) )

   info = reader.channelInfo(frame, reader.channelCount - 1)
   print( str(info) )

if __name__ == "__main__":
   if len( sys.argv ) < 2:
      print( 'Usage: raw.py [raw-file]')
      sys.exit(0)

   import logging
   logging.basicConfig(
      level=logging.DEBUG,
      format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
      datefmt='%Y-%m-%d %H:%M:%S' )
   logger = logging.getLogger( 'CameraRawReader' )

   try:
      logger.info( "+++ Start of run" )
      import sys
      main()
   except Exception as detail:
      logger.critical( '{0}'.format( detail ) )
   finally:
      logger.info( "--- End of run" )

#
# end of file
#
