#
# camera/formats/util.py
#
# Utility functions to read binary data.
#
# This file is part of the Python package to handle Camera data (CAMPP).
#
# Copyright (c) 2009-2012, Universiteit Leiden
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# - Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# - Neither the name of Universiteit Leiden nor the names of its contributors
#   may be used to endorse or promote products derived from this software
#   without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
#
# Authors: Martin Moene (original), Sander Roobol
#
# $Id: util.py 649 2012-02-27 13:41:59Z moene $
#

import struct

def readBinaryHeader(f, description):
    ret = {}
    for name, type in description:
        ret[name] = type(f)
    return ret

def _readBinary(f, xtype, xsizeof):
   """Read an binary item at the current file position."""
   return struct.unpack(xtype, f.read(xsizeof))[0]

def readDouble(f):
   """Read a double at the current file position."""
   return _readBinary(f, '<d', 8)

def readInt32(f):
   """Read a 32-bit int at the current file position."""
   return _readBinary(f, '<l', 4)

def readInt16(f):
   """Read a 16-bit int at the current file position."""
   return _readBinary(f, '<h', 2)

def readUInt32(f):
   """Read an unsigned 32-bit int at the current file position."""
   return _readBinary(f, '<L', 4)

def readUInt16(f):
   """Read an unsigned 16-bit int at the current file position."""
   return _readBinary(f, '<H', 2)

def readByte(f):
   """Read a byte at the current file position."""
   return f.read(1)

def readZString(f):
   """Read a nul-terminated string at the current file position.
   Ok for Python 2.6 and Python 3.
   """
   s = b''
   b = readByte(f)
   while b'\x00' != b:
      if b'\x20' <= b <= b'\x7f':
         s += b
      else:
         s += b'?'
      b = readByte(f)
   return s.decode("ascii")   # str(s, encoding='ascii')

def readCString(f):
    """Read CArchive-style string: data prepended by 8-bit integer indicating the length."""
    length = ord(f.read(1))
    return f.read(length).decode("ascii")

def readObjName(f):
    """Read CArchive object name: a string prepended by its length as an 16-byte integer"""
    length = readUInt16(f)
    return f.read(length).decode("ascii")

# some convenient aliases for use with readBinaryHeader
DOUBLE = readDouble
INT32 = readInt32
INT16 = readInt16
UINT32 = readUInt32
UINT16 = readUInt16
CHAR = readByte
ZSTR = readZString
CSTR = readCString
OBJNAME = readObjName

def bitcount(x):
   """the number of set bits, see http://wiki.python.org/moin/BitManipulation"""
   n = int(0)
   while x:
      n += 1
      x &= x - 1  # clear the least significant bit set
   return n
